/*
 * libarc.c -- libarc archive handler
 * (C)Copyright 1999,2000 by TAJIRI Yasuhiro
 * This file is part of Enfle.
 *
 */

#include "enfle.h"
#include "mfile.h"
#include "libenflearc.h"
#include "utils.h"
#include "plugin.h"

#define MAX_DENSITY 5
#define INIT_TABLE_SIZE 1000

extern struct URL_module URL_module_file;
extern struct URL_module URL_module_http;
int libarc_archive_select(Archive *ar, int index);
int libarc_archive_read(Archive *ar, unsigned char *buf, int size);
int libarc_archive_close(Archive *ar);
int libarc_archive_seek(Archive *ar, long seek,int whence);
int libarc_archive_tell(Archive *ar);

int
#ifdef PIC
get_plugininfo(PluginInfo *p)
#else
archiver_pak_get_plugininfo(PluginInfo *p)
#endif
{
  p->version = 1;
  p->type = _Archiver;
  p->pluginname = "Archiver Plugin via libarc version 0.2";
  p->pluginshortname = FORMAT_NAME;
  p->author = "TAJIRI Yasuhiro";
  p->dlhandle = NULL; /* set by plugin_load */
  p->functions.archiver.open = libarc_archive_open;
  return 1;
}

int
libarc_archive_open(Archive *ar)
{
    ArchiveFileList* archive_file_list = NULL;
    int nfiles=1;
    url_add_modules(
	&URL_module_file,
	NULL);
    archive_file_list = make_archive_list(
        archive_file_list,1, &(ar->filename));
    if(archive_file_list != NULL)
    {
        char** new_files;
        int    new_nfiles;

        new_nfiles = nfiles;
        new_files = expand_archive_names(archive_file_list,
                                         &new_nfiles, &(ar->filename));
        if(new_files != NULL && archive_file_list != NULL)
        {
            LIBARC_info * info;
            if ((ar->info = calloc(1, sizeof(LIBARC_info))) == NULL) {
                fprintf(stderr, "No enough memory for info\n");
                goto error;
            }
            info = (LIBARC_info *)(ar->info);
            ar->nfiles=info->nfile  = new_nfiles;
            info->files  = new_files;
            info->url    = NULL;
            info->mem    = NULL;
            info->archive_file_list = archive_file_list;
        }
        else goto error;
    }
    else goto error;
    if (ar->nfiles) {
        ar->format = FORMAT_NAME;
        ar->select = libarc_archive_select;
        ar->close  = libarc_archive_close;
        ar->read   = libarc_archive_read;
        ar->seek   = libarc_archive_seek;
        ar->tell   = libarc_archive_tell;
        return 1;
    }
  error:
    free(ar->info);
    return 0;
}

int
libarc_archive_select(Archive *ar, int index)
{
    LIBARC_info *info = ar->info;

    if(ar==NULL)
        return 0;
    if (index < 0)
        return 0;
    if(info->url != NULL)
        url_close(info->url);
    if(info->mem ==NULL)
      info->mem = mopen(10000);
    else
        mseek(info->mem,0,SEEK_SET);
    info->url = archive_file_extract_open(info->archive_file_list,
                                          info->files[index]);
    if(info->url == NULL)
        info->url = url_open(info->files[index]);
    if(info->url == NULL)
    {
        fprintf(stderr, "Can't open: %s\n", info->files[index]);
        return 0;
    }
    ar->size=0;
    
    while(1){
        unsigned char buf[1000];
        int size=url_read(info->url,buf,sizeof(buf));
        ar->size+=size;
        mwrite(buf,1,size,info->mem);
        if(size != sizeof(buf))
            break;
    }
    mseek(info->mem,0,SEEK_SET);
    return 1;
}

int
libarc_archive_seek(Archive *ar, long seek,int whence)
{
    return mseek(((LIBARC_info*)(ar->info))->mem, seek, whence);
}

int
libarc_archive_tell(Archive *ar)
{
    return mtell(((LIBARC_info*)(ar->info))->mem);
}
        

int
libarc_archive_read(Archive *ar, unsigned char *buf, int size)
{
    return mread(buf,1,size,((LIBARC_info*)(ar->info))->mem);
}
        
int
libarc_archive_close(Archive *ar)
{
    mclose(((LIBARC_info*)(ar->info))->mem);
    free(ar->info);
    ar->info=NULL;
    return 1;
}
