/*
 * render.c -- enfle rendering engine
 * (C)Copyright 1998, 99 by Hiroshi Takekawa
 * This file is part of Enfle.
 *
 * Last Modified: Sat Dec 11 02:42:40 1999.
 * $Id: render.c,v 1.17 1999/12/23 15:16:27 sian Exp $
 *
 * Enfle is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Enfle is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include "enfle.h"
#include "cmrender.h"
#include "quantize.h"

XImage *
render_image(Info *info, Image *p, int f)
{
  XImage *ximage;
  int width = p->width;
  int height = p->height;
  int depth = info->depth;
  int i, j;
  unsigned char *xdata;

  if (!f) {
    info->xwidth = width + p->left;
    info->xheight = height + p->top;
    if (info->swidth)
      info->xwidth = info->mswidth;
    if (info->sheight)
      info->xheight = info->msheight;

    ximage = XCreateImage(info->disp, info->visual, depth, ZPixmap, 0, NULL,
			  info->xwidth, info->xheight, 8, 0);
  
    info->xdata_size = info->xwidth * ((ximage->bits_per_pixel + 7) >> 3) * info->xheight;
    if (info->xdata != NULL)
      free(info->xdata);
    if ((info->xdata = calloc(1, info->xdata_size)) == NULL) {
      fprintf(stderr, "No enough memory xdata(%dx%d, depth = %d) %d bytes\n",
	      width, height, info->depth, info->xdata_size);
      exit(1);
    }
  } else
    ximage = XCreateImage(info->disp, info->visual, depth, ZPixmap, 0, NULL,
			  info->xwidth, info->xheight, 8, 0);

  ximage->data = xdata = info->xdata;

#ifdef DEBUG
  fprintf(stderr, "type = %d default depth = %d\n", p->type, depth);
  fprintf(stderr, "(XImage's)bits_per_pixel = %d byte_order: %s bitmap_bit_order: %s\n",
	  ximage->bits_per_pixel,
	  (ximage->byte_order == LSBFirst) ? "LSBFirst" : "MSBFirst",
	  (ximage->bitmap_bit_order == LSBFirst) ? "LSBFirst" : "MSBFirst");
#endif
  ximage->byte_order = (depth == 24 && (p->type == _RGB24 || p->type == _RGB16)) ? MSBFirst : LSBFirst;

  /* expansion */
  switch (p->type) {
  case _MONO:
    if (depth != 1 && !image_expand(p, _INDEX)) {
      fprintf(stderr, "No enough memory for expand %d bytes\n", width * height);
      exit(1);
    }
    break;
  case _GRAY:
    if (!image_expand(p, _INDEX)) {
      fprintf(stderr, "Internal error: Error occurred during image_expand()\n");
      exit(1);
    }
    break;
  case _RGB16:
#ifdef DEBUG
    fprintf(stderr, "_RGB16\n");
#endif
    if (depth != 16 && depth != 15 && !image_expand(p, _RGB24)) {
      fprintf(stderr, "No enough memory for expand %d bytes\n", width * height * 3);
      exit(1);
    }
    p->ncolors = 65536;
    break;
  case _RGB24:
#ifdef DEBUG
    fprintf(stderr, "_RGB24\n");
#endif
    break;
  case _INDEX:
#ifdef DEBUG
    fprintf(stderr, "_INDEX\n");
#endif
    break;
  default:
    break;
  }

  /* reduction */
  if (!info->if_quantize) {
    switch (p->type) {
    case _RGB24:
      if (depth == 16 && !image_reduce(p, _RGB16)) {
	fprintf(stderr, "render_image: FATAL: image_reduce: returns 0\n");
	exit(1);
      } else if (depth == 15 && !image_reduce(p, _RGB15)) {
	fprintf(stderr, "render_image: FATAL: image_reduce: returns 0\n");
	exit(1);
      }
      break;
    default:
    }
  }

  /* rendering */
  if (!info->if_quantize) {
    switch (p->type) {
    case _MONO:
    case _GRAY:
      /* cannot reach here */
      fprintf(stderr, "Internal error, cannot rendering\n");
      exit(1);
    case _RGB16:
      if (depth != 16 && depth != 15) {
	/* cannot reach here */
	fprintf(stderr, "Internal error, cannot rendering\n");
	exit(1);
      }
      if (ximage->bits_per_pixel == 16) {
	/* acceleration code */
	free(xdata);
	info->xdata = NULL;
	ximage->data = xdata = p->image;
	/* memcpy(xdata, p->image, width * height * 2); */
      } else {
	fprintf(stderr, "I have no code for bits_per_pixel = %d\n", ximage->bits_per_pixel);
	exit(1);
      }
      break;
    case _RGB24:
      if (depth != 24) {
	fprintf(stderr, "I have no code to render _RGB24 for depth %d\n", depth);
	exit(1);
      }
      if (ximage->bits_per_pixel == 24) {
	/* acceleration code */
	free(xdata);
	info->xdata = NULL;
	ximage->data = xdata = p->image;
	/* memcpy(xdata, p->image, width * height * 3); */
      } else if (ximage->bits_per_pixel == 32) {
	unsigned char *s = p->image;
	unsigned char *d = xdata;

	for (i = 0; i < width * height; i++) {
	  d[0] = 0;
	  d[1] = *s++;
	  d[2] = *s++;
	  d[3] = *s++;
	  d += 4;
	}
      } else {
	fprintf(stderr, "I have no code for bits_per_pixel = %d\n", ximage->bits_per_pixel);
	exit(1);
      }
      break;
    default:
      break;
    }
  } else {
    if (p->ncolors > 256) {
      /* image quantization */
      unsigned char *q, pal[256 * 3];

      if ((q = quantize(p->image, pal, width, height, 8)) == NULL)
	return NULL;

      for (i = 0; i < 256; i++)
	for (j = 0; j < 3; j++)
	  p->colormap[i][j] = pal[i * 3 + j];
      free(p->image);
      p->image = q;
      p->type = _INDEX;
      p->bytes_per_line = p->width;
      p->ncolors = 256;
    }
  }

  if (p->type == _INDEX) {
    int f;

    /* colormapped image rendering */
    f = render_image_colormap(info, p, xdata, ximage->bits_per_pixel);
    if (f) {
      fprintf(stderr, "enfle: render_image: FATAL: %s\n", cmrender_err[f]);
      exit(1);
    }
  }

#ifdef SHAPE
  if (p->transparent_disposal == _SHAPE) {
    XImage *xmask;
    GC gc;
    XGCValues gcv;

    if (p->mask == NULL)
      image_create_mask(p);

    xmask = XCreateImage(info->disp, info->visual, 1, XYBitmap, 0, p->mask,
			 p->width, p->height, 8, 0);
    xmask->byte_order = MSBFirst;
    xmask->bitmap_bit_order = MSBFirst;
    p->pixmask = XCreatePixmap(info->disp, info->win, p->left + p->width, p->top + p->height, 1);
    gcv.background = 1;
    gcv.foreground = 0;
    gc = XCreateGC(info->disp, p->pixmask, GCBackground | GCForeground, &gcv);
    XFillRectangle(info->disp, p->pixmask, gc, 0, 0, p->left + p->width, p->top + p->height);
    XPutImage(info->disp, p->pixmask, gc, xmask, 0, 0, p->left, p->top, p->width, p->height);
    XFreeGC(info->disp, gc);
    XShapeCombineMask(info->disp, info->win, ShapeBounding, 0, 0, p->pixmask, ShapeSet);
    XDestroyImage(xmask);
    p->mask = NULL;
  }
#endif

  return ximage;
}
